(********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis SOURCE CODE IS (C) COPYRIGHT 1994-2002             *
 * by the XIPHOPHORUS Company http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: simple example decoder
 last mod: $Id: decoder_example.c,v 1.27 2002/07/12 15:07:52 giles Exp $

 ********************************************************************)

(* Takes a vorbis bitstream from input file and writes raw stereo PCM to
   output file. Decodes simple and chained OggVorbis files from beginning
   to end.  Vorbisfile.a is somewhat more complex than the code below.  *)

{ =============================================

  Delphi conversion: Aleksandr "Lake" Shamray
		     http://lakeofsoft.com/

		     last modified: 05 November 2002

 =============================================}

{$APPTYPE CONSOLE}

{$Q-,B-,O+,R-,I-}

program
  decoder_example;

uses
  unaVorbisAPI;

const
  convsize = 4096;

type
  int = integer;

  pCharArray = ^tCharArray;
  tCharArray = array[0.. maxInt div sizeOf(pChar) - 1] of pChar;

var
  convbuffer: array[0 .. convsize - 1] of ogg_int16_t;	//* take 8k out of the data segment, not the stack */
  fileIn, fileOut: file;

// --  --
procedure run();
var
  oy: tOgg_sync_state; //* sync and verify incoming physical bitstream */
  os: tOgg_stream_state; //* take physical pages, weld into a logical stream of packets */
  og: tOgg_page; //* one Ogg bitstream page.  Vorbis packets are inside */
  op: tOgg_packet; //* one raw packet of data for decode */

  vi: tVorbis_info; //* struct that stores all the static vorbis bitstream settings */
  vc: tVorbis_comment; //* struct that stores all the bitstream user comments */
  vd: tVorbis_dsp_state; //* central working state for the packet->PCM decoder */
  vb: tVorbis_block; //* local working space for packet->PCM decode */

  comments: pCharArray;

  buffer: pointer;
  readCount, writeCount: int;
  inCount, outCount: int64;
  _convsize: int;

  pcm: pSingleSamples;
  samples: int;
  j, clipflag, bout: int;
  mono: pSingleArray;
  val: int;

  eos, i, result: int;
begin
  inCount := 0;
  outCount := 44;

  //********** Decode setup ************/
  ogg_sync_init(oy); //* Now we can read pages */

  while (true) do begin //* we repeat if the bitstream is chained */
    //
    eos := 0;

    (* grab some data at the head of the stream.  We want the first page
       (which is guaranteed to be small and only contain the Vorbis
       stream initial header) We need the first page to get the stream
       serialno. *)

    //* submit a 4k block to libvorbis' Ogg layer */
    buffer := ogg_sync_buffer(oy, convsize);
    blockRead(fileIn, buffer^, convsize, readCount);
    inc(inCount, readCount);
    ogg_sync_wrote(oy, readCount);

    //* Get the first page. */
    if (1 <> ogg_sync_pageout(oy, og)) then begin
      //* have we simply run out of data?  If so, we're done. */
      if (readCount < convsize) then
	break;

      //* error case.  Must not be Vorbis data */
      writeLn('Input does not appear to be an Ogg bitstream.');
      exit;
    end;

    //* Get the serial number and set up the rest of decode. */
    //* serialno first; use it to set up a logical stream */
    ogg_stream_init(os, ogg_page_serialno(og));

    //* extract the initial header from the first page and verify that the Ogg bitstream is in fact Vorbis data */

    (* I handle the initial header first instead of just having the code
       read all three Vorbis headers at once because reading the initial
       header is an easy way to identify a Vorbis bitstream and it's
       useful to see that functionality seperated out. *)

    vorbis_info_init(vi);
    vorbis_comment_init(vc);
    if (0 > ogg_stream_pagein(os, og)) then begin
      //* error; stream version mismatch perhaps */
      writeLn('Error reading first page of Ogg bitstream data.');
      exit;
    end;

    if (1 <> ogg_stream_packetout(os, op)) then begin
      //* no page? must not be vorbis */
      writeLn('Error reading initial header packet.');
      exit;
    end;

    if (0 > vorbis_synthesis_headerin(vi, vc, op)) then begin
      //* error case; not a vorbis header */
      writeLn('This Ogg bitstream does not contain Vorbis audio data.');
      exit;
    end;

    (* At this point, we're sure we're Vorbis.  We've set up the logical
       (Ogg) bitstream decoder.  Get the comment and codebook headers and
       set up the Vorbis decoder *)

    (* The next two packets in order are the comment and codebook headers.
       They're likely large and may span multiple pages.  Thus we reead
       and submit data until we get our two pacakets, watching that no
       pages are missing.  If a page is missing, error out; losing a
       header page is the only place where missing data is fatal. *)

    i := 0;
    while (2 > i) do begin
      while (2 > i) do begin
	//
	result := ogg_sync_pageout(oy, og);
	if (0 = result) then
	  break; //* Need more data */

	//* Don't complain about missing or corrupt data yet.  We'll catch it at the packet output phase */
	if (1 = result) then begin
	  ogg_stream_pagein(os, og); //* we can ignore any errors here as they'll also become apparent at packetout */

	  while (2 > i) do begin
	    //
	    result := ogg_stream_packetout(os, op);
	    if (0 = result) then
	      break;
	    //
	    if (0 > result) then begin
	      //* Uh oh; data at some point was corrupted or missing! We can't tolerate that in a header.  Die. */
	      writeLn('Corrupt secondary header.  Exiting.');
	      exit;
	    end;
	    vorbis_synthesis_headerin(vi, vc, op);
	    inc(i);
	  end;
	end;
      end;

      //* no harm in not checking before adding more */
      buffer := ogg_sync_buffer(oy, convsize);
      blockRead(fileIn, buffer^, convsize, readCount);
      inc(inCount, readCount);
      //
      if ((0 = readCount) and (2 > i)) then begin
	writeLn('End of file before finding all Vorbis headers!');
	exit;
      end;
      //
      ogg_sync_wrote(oy, readCount);
    end;

    //* Throw the comments plus a few lines about the bitstream we're decoding */
    begin
      comments := vc.user_comments;
      //
      i := 0;
      while (nil <> comments[i]) do begin
	writeLn(comments[i]);
	inc(i);
      end;
      //
      writeLn(#13#10'Bitstream is ', vi.channels, ' channel, ', vi.rate, ' Hz');
      writeLn('Encoded by: ', vc.vendor, #13#10);
    end;

    _convsize := convsize div vi.channels;

    //* OK, got and parsed all three headers. Initialize the Vorbis packet->PCM decoder. */
    vorbis_synthesis_init(vd, vi); //* central decode state */
    vorbis_block_init(vd, vb);     (* local state for most of the decode
				       so multiple block decodes can
				       proceed in parallel.  We could init
				       multiple vorbis_block structures
				       for vd here *)

    //* The rest is just a straight decode loop until end of stream */
    while (0 = eos) do begin
      while(0 = eos) do begin
	//
	result := ogg_sync_pageout(oy, og);
	if (0 = result) then
	  break; //* need more data */

	if (0 > result) then //* missing or corrupt data at this page position */
	  writeLn('Corrupt or missing data in bitstream; continuing...')
	else begin
	  ogg_stream_pagein(os, og); //* can safely ignore errors at this point */

	  while (true) do begin
	    //
	    result := ogg_stream_packetout(os, op);
	    if (0 = result) then
	      break; //* need more data */
	    //
	    if (0 > result) then //* missing or corrupt data at this page position */
	      //* no reason to complain; already complained above */

	    else begin
	      //* we have a packet.  Decode it */
	      if (0 = vorbis_synthesis(vb, op)) then //* test for success! */
		vorbis_synthesis_blockin(vd, vb);

	      (*
	      **pcm is a multichannel float vector.  In stereo, for
	      example, pcm[0] is left, and pcm[1] is right.  samples is
	      the size of each channel.  Convert the float values
	      (-1.<=range<=1.) to whatever PCM format and write it out *)

	      repeat
		samples := vorbis_synthesis_pcmout(vd, pcm);
		if (1 > samples) then
		  break;
		//
		clipflag := 0;
		if (samples <_convsize) then
		  bout := samples
		else
		  bout := _convsize;

		//* convert floats to 16 bit signed ints (host order) and interleave */
		i := 0;
		while (i < vi.channels) do begin
		  //
		  mono := pcm[i];
		  j :=0;
		  while (j < bout) do begin
		    //
		    val := trunc(mono[j] * 32767);

		    //* optional dither */
		    //int val=mono[j]*32767.f+drand48()-0.5f;

		    //* might as well guard against clipping */
		    if (val > 32767) then begin
		      val := 32767;
		      clipflag := 1;
		    end;
		    //
		    if (val < -32768) then begin
		      val := -32768;
		      clipflag := 1;
		    end;
		    //
		    convbuffer[i + vi.channels * j] := val;
		    inc(j);
		  end;
		  //
		  inc(i);
		end;

		if (0 <> clipflag) then
		  writeLn('Clipping in frame ', vd.sequence, '                               ');

		//
		blockWrite(fileOut, convbuffer, 2 * vi.channels * bout, writeCount);
		inc(outCount, writeCount);
		//
		vorbis_synthesis_read(vd, bout); //* tell libvorbis how  many samples we actually consumed */
	      until (false);
	    end;
	  end;
	  //
	  if (0 <> ogg_page_eos(og)) then
	    eos := 1;
	end;
      end;

      if (0 = eos) then begin
	//
	buffer := ogg_sync_buffer(oy, convsize);
	blockRead(fileIn, buffer^, convsize, readCount);
	inc(inCount, readCount);
	//
	ogg_sync_wrote(oy, readCount);
	if (0 = readCount) then
	  eos := 1;
      end;

      //
      write('input: ', inCount, ';   output: ', outCount, '      '#13);
    end;

    //* clean up this logical bitstream; before exit we see if we're followed by another [chained] */

    ogg_stream_clear(os);

    //* ogg_page and ogg_packet structs always point to storage in libvorbis.  They're never freed or manipulated directly */

    vorbis_block_clear(vb);
    vorbis_dsp_clear(vd);
    vorbis_comment_clear(vc);
    vorbis_info_clear(vi);  //* must be called last */
  end;

  //* OK, clean up the framer */
  ogg_sync_clear(oy);

  writeLn('Done');
end;

// -- main --

const
  c_wav44Khz16stereo_header: array[1..44] of byte = (
    $52, $49, $46, $46, $34, $73, $B8, $03, $57, $41, $56, $45, $66, $6D, $74, $20,
    $10, $00, $00, $00, $01, $00, $02, $00, $44, $AC, $00, $00, $10, $B1, $02, $00,
    $04, $00, $10, $00, $64, $61, $74, $61, $10, $73, $B8, $03);

begin
  if (paramCount < 2) then begin
    writeLn('must specify input (ogg) and output (wav) file names.');
    exit;
  end
  else begin
    //
    if (vorbis_load_library(cunav_dll_ogg) and vorbis_load_library(cunav_dll_vorbis)) then begin
      //
      assignFile(fileIn, paramStr(1));
      reset(fileIn, 1);	// record size = 1
      //
      assignFile(fileOut, paramStr(2));
      rewrite(fileOut, 1);	// record size = 1
      // cheat: write out "false" WAV header for 44 kHz; 16 bits stereo PCM stream; 5 min. 53 sec.
      blockWrite(fileOut, c_wav44Khz16stereo_header, sizeOf(c_wav44Khz16stereo_header));
      //
      try
	run();
      finally
	//
	closeFile(fileIn);
	closeFile(fileOut);
	//
	vorbis_unload_library(cunav_dll_ogg);
	vorbis_unload_library(cunav_dll_vorbis);
      end;
    end
    else
      writeLn('no vorbis library was found.');
  end;
end.

